SIGNAL_HOST_HOST_MSG = "EvHostHandMsg"
SIGNAL_HOST_HAND_MSG = "EvHostHandMsg"
SIGNAL_HAND_HOST_MSG = "EvHandHostMsg"
SIGNAL_HAND_PROC_MSG = "EvHandProcMsg"
SIGNAL_PROC_HAND_MSG = "EvProcHandMsg"
SIGNAL_PROC_NSTK_MSG = "EvProcNstkMsg"
SIGNAL_NSTK_PROC_MSG = "EvNstkProcMsg"
SIGNAL_PEER_NSTK_MSG = "EvPeerNstkMsg"

SIGNAL_MEMORY_FREED = "EvMemoryFreed"
